import os
from copy import deepcopy
from csv import DictReader
from importlib import import_module
from time import time
from typing import List

from .automatix import Automatix
from .command import Command
from .config import arguments, CONFIG, get_script, LOG, update_script_from_row, collect_vars, SCRIPT_FIELDS

if CONFIG.get('logging_lib'):
    log_lib = import_module(CONFIG.get('logging_lib'))
    init_logger = log_lib.init_logger
else:
    from .logger import init_logger

if CONFIG.get('bundlewrap'):
    from bundlewrap.repo import Repository
    from .bundlewrap import BWCommand

    CONFIG['bw_repo'] = Repository(repo_path=os.environ.get('BW_REPO_PATH'))
    cmdClass = BWCommand
else:
    cmdClass = Command


def main():
    starttime = time()
    args = arguments()
    init_logger(name=CONFIG['logger'], debug=args.debug)

    script = get_script(args=args)

    batch_items: List[dict] = [{}]
    if args.vars_file:
        with open(args.vars_file) as csvfile:
            batch_items = list(DictReader(csvfile))
        LOG.notice('Detected batch processing from CSV file.')

    for i, row in enumerate(batch_items, start=1):
        script_copy = deepcopy(script)
        update_script_from_row(row=row, script=script_copy, index=i)

        variables = collect_vars(script_copy)

        auto = Automatix(
            script=script_copy,
            variables=variables,
            config=CONFIG,
            cmd_class=cmdClass,
            script_fields=SCRIPT_FIELDS,
        )

        auto.run(args=args)

    if 'AUTOMATIX_TIME' in os.environ:
        LOG.info(f'The Automatix script took {round(time() - starttime)}s!')



import argparse
import logging
import os
from collections import OrderedDict

import yaml

yaml.warnings({'YAMLLoadWarning': False})


def read_yaml(yamlfile: str) -> dict:
    with open(yamlfile) as file:
        return yaml.load(file.read())


SCRIPT_FIELDS = OrderedDict()
SCRIPT_FIELDS['systems'] = 'Systems'
SCRIPT_FIELDS['vars'] = 'Variables'

CONFIG = {
    'script_dir': '~/automatix-config',
    'constants': {},
    'encoding': os.getenv('ENCODING', 'utf-8'),
    'import_path': '.',
    'ssh_cmd': 'ssh {hostname} sudo ',
    'remote_tmp_dir': 'automatix_tmp',
    'logger': 'automatix',
    'bundlewrap': False,
    'teamvault': False,
}

configfile = os.path.expanduser(os.getenv('AUTOMATIX_CONFIG', '~/.automatix.cfg.yaml'))
if os.path.isfile(configfile):
    CONFIG.update(read_yaml(configfile))

if os.getenv('AUTOMATIX_SCRIPT_DIR'):
    CONFIG['script_dir'] = os.getenv('AUTOMATIX_SCRIPT_DIR')

LOG = logging.getLogger(CONFIG['logger'])

SCRIPT_PATH = os.path.expanduser(CONFIG['script_dir'])

if CONFIG['teamvault']:
    import bwtv

    SCRIPT_FIELDS['secrets'] = 'Secrets'


    class UnknownSecretTypeException(Exception):
        pass


def arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description='Automation wrapper for bash and python commands.',
        epilog='Explanations and README at https://github.com/seibert-media/automatix',
    )
    parser.add_argument(
        'scriptfile',
        help='Path to scriptfile (yaml), use " -- " if needed to delimit this from argument fields',
    )
    for field in SCRIPT_FIELDS.keys():
        parser.add_argument(
            f'--{field}',
            nargs='*',
            help=f'Use this to set {field} without adding them to the script or to overwrite them. '
                 f'You can specify multiple {field} like: --{field} v1=string1 v2=string2 v3=string3',
        )
    parser.add_argument(
        '--vars-file',
        help='Path to a CSV file containing variables for batch processing',
    )
    parser.add_argument(
        '--print-overview', '-p',
        action='store_true',
        help='Just print command pipeline overview with indices then exit without executing the commandline. '
             'Note that the *always pipeline* will be executed beforehand.',
    )
    parser.add_argument(
        '--jump-to', '-j',
        help='Jump to step instead of starting at the beginning',
        default=0,
    )
    parser.add_argument(
        '--interactive', '-i',
        action='store_true',
        help='Confirm actions before executing',
    )
    parser.add_argument(
        '--force', '-f',
        action='store_true',
        help='try always to proceed (except manual steps), even if errors occur (no retries)'
    )
    parser.add_argument(
        '--debug', '-d',
        action='store_true',
        help='activate debug log level',
    )
    return parser.parse_args()


def _overwrite(script: dict, key: str, data: str):
    script.setdefault(key, {})
    for item in data:
        k, v = item.split('=')
        script[key][k] = v


def get_script(args: argparse.Namespace) -> dict:
    file = args.scriptfile
    if not os.path.isfile(args.scriptfile):
        file = f'{SCRIPT_PATH}/{args.scriptfile}'

    script = read_yaml(file)

    for field in SCRIPT_FIELDS.keys():
        if vars(args).get(field):
            _overwrite(script=script, key=field, data=vars(args)[field])

    return script


def collect_vars(script: dict) -> dict:
    var_dict = script.get('vars', {})
    script['vars'] = var_dict  # just for the case it was empty
    if CONFIG['teamvault']:
        for key, secret in script.get('secrets', {}).items():
            sid, field = secret.split('_')
            if field == 'password':
                var_dict[key] = bwtv.password(sid)
            elif field == 'username':
                var_dict[key] = bwtv.username(sid)
            elif field == 'file':
                var_dict[key] = bwtv.file(sid)
            else:
                raise UnknownSecretTypeException(field)
    for syskey, system in script.get('systems', {}).items():
        var_dict[f'system_{syskey}'] = system
    return var_dict


def update_script_from_row(row: dict, script: dict, index: int):
    if not row:
        return
    try:
        script['name'] += f" ({index} {row.pop('label')})"
    except KeyError:
        script['name'] += f" ({index})"
    for key, value in row.items():
        assert len(key.split(':')) == 2, \
            'First row in CSV must contain "label" or the field name and key seperated by colons' \
            ' like "label,systems:mysystem,vars:myvar".'
        key_type, key_name = key.split(':')
        assert key_type in SCRIPT_FIELDS.keys(), \
            f'First row in CSV: Field name is \'{key_type}\', but has to be one of {list(SCRIPT_FIELDS.keys())}.'
        script[key_type][key_name] = value



from logging import Logger


class PipelineEnvironment:
    def __init__(self, name: str, config: dict, systems: dict, variables: dict, imports: list, LOG: Logger):
        self.name = name
        self.config = config
        self.systems = systems
        self.vars = variables
        self.imports = imports
        self.LOG = LOG



import os
import re
import subprocess
import traceback

from shlex import quote
from time import time
from typing import Tuple

from .environment import PipelineEnvironment

PERSISTENT_VARS = {}


class Command:
    def __init__(self, pipeline_cmd: dict, index: int, env: PipelineEnvironment):
        self.pipeline_cmd = pipeline_cmd
        self.index = index
        self.env = env

        for key, value in pipeline_cmd.items():
            self.orig_key = key
            self.condition_var, self.assignment_var, self.key = parse_key(key=key)
            if isinstance(value, dict):
                # We need this workaround because the yaml lib returns a dictionary instead of a string,
                # if there is nothing but a variable in the command. Alternative is to use quotes in the script yaml.
                self.value = f'{{{next(iter(value))}}}'
            else:
                self.value = value
            break  # There should be only one entry in pipeline_cmd

    def get_type(self):
        if self.key == 'local':
            return 'local'
        if self.key == 'manual':
            return 'manual'
        if self.key == 'python':
            return 'python'
        if re.search('remote@', self.key):
            return 'remote'
        raise UnknownCommandException(f'Command type {self.key} is not known.')

    def get_system(self):
        if self.get_type() == 'remote':
            return self.env.systems[re.search(r'remote@(.*)', self.key).group(1)]
        return 'localhost'

    def get_resolved_value(self):
        variables = self.env.vars.copy()
        for key, value in variables.items():
            if not isinstance(value, str):
                continue
            file_match = re.match(r'FILE_(.*)', value)
            if file_match:
                with open(os.path.expanduser(file_match.group(1))) as file:
                    variables[key] = file.read().strip()

        for key, value in self.env.config['constants'].items():
            variables[f'const_{key}'] = value
        return self.value.format(**variables)

    def execute(self, interactive: bool = False, force: bool = False):
        self.env.LOG.notice(f'\n({self.index}) [{self.orig_key}]: {self.get_resolved_value()}')
        return_code = 0

        if self.condition_var is not None and not bool(self.env.vars.get(self.condition_var, False)):
            self.env.LOG.info(f'Skip command, because condition variable "{self.condition_var}" evolves to False')
            return

        if self.get_type() == 'manual' or interactive:
            self.env.LOG.debug('Ask for user interaction.')
            answer = input(f'[MS] Proceed? (p: proceed (default), s: skip, a: abort)\n')
            if answer == 's':
                return
            if answer == 'a':
                raise AbortException(1)

        steptime = time()

        if self.get_type() == 'local':
            return_code = self._local_action()
        if self.get_type() == 'python':
            return_code = self._python_action()
        if self.get_type() == 'remote':
            return_code = self._remote_action()

        if 'AUTOMATIX_TIME' in os.environ:
            self.env.LOG.info(f'(command execution time: {round(time()-steptime)}s)')

        if return_code != 0:
            self.env.LOG.error(f'>> {self.env.name} << Command ({self.index}) failed with return code {return_code}.')
            if force:
                return
            err_answer = input('[CF] What should I do? (p: proceed (default), r: retry, a: abort)\n')
            if err_answer == 'r':
                return self.execute(interactive)
            if err_answer == 'a':
                raise AbortException(return_code)

    def _local_action(self) -> int:
        cmd = self._build_command(path=self.env.config['import_path'])
        try:
            return self._run_local_command(cmd=cmd)
        except KeyboardInterrupt:
            self.env.LOG.info('Abort command by user key stroke. Exit code is set to 130.')
            return 130

    def _generate_python_vars(self):
        return {'vars': self.env.vars}

    def _python_action(self) -> int:
        cmd = self.get_resolved_value()
        locale_vars = self._generate_python_vars()
        locale_vars.update(PERSISTENT_VARS)
        self.env.LOG.debug(f'locals:\n {locale_vars}')
        try:
            self.env.LOG.debug(f'Run python command: {cmd}')
            if self.assignment_var:
                exec(f'vars["{self.assignment_var}"] = {cmd}', globals(), locale_vars)
                self.env.LOG.info(f'Variable {self.assignment_var} = {self.env.vars[self.assignment_var]}')
            else:
                exec(cmd, globals(), locale_vars)
            return 0
        except KeyboardInterrupt:
            self.env.LOG.info('Abort command by user key stroke. Exit code is set to 130.')
            return 130
        except Exception:
            self.env.LOG.error(traceback.format_exc())
            return 1

    def _get_remote_hostname(self):
        return self.get_system()

    def _remote_action(self) -> int:
        hostname = self._get_remote_hostname()

        ssh_cmd = self.env.config["ssh_cmd"].format(hostname=hostname)
        remote_cmd = self.get_resolved_value()
        prefix = ''
        if self.env.imports:
            prefix = f'tar -C {self.env.config["import_path"]} -cf - ' + ' '.join(self.env.imports) + ' | '
            remote_cmd = f'mkdir {self.env.config["remote_tmp_dir"]};' \
                         f' tar -C {self.env.config["remote_tmp_dir"]} -xf -;' \
                         f' {self._build_command(path=self.env.config["remote_tmp_dir"])}'

        cmd = f'{prefix}{ssh_cmd}{quote("bash -c " + quote(remote_cmd))}'

        try:
            exitcode = self._run_local_command(cmd=cmd)
        except KeyboardInterrupt:
            self.env.LOG.info('Abort command by user key stroke. Exit code is set to 130.')
            exitcode = 130

            try:
                ps_pids = self.get_remote_pids(hostname=hostname, cmd=self.get_resolved_value())
                while ps_pids:
                    self.env.LOG.notice(
                        'Remote command seems still to be running! Found PIDs: {}'.format(','.join(ps_pids))
                    )
                    answer = input(
                        '[RR] What should I do? '
                        '(i: send SIGINT (default), t: send SIGTERM, k: send SIGKILL, p: do nothing and proceed) \n')

                    if answer == 'p':
                        break
                    elif answer == 't':
                        signal = 'TERM'
                    elif answer == 'k':
                        signal = 'KILL'
                    else:
                        signal = 'INT'
                    for pid in ps_pids:
                        kill_cmd = f'{ssh_cmd} kill -{signal} {pid}'
                        self.env.LOG.info(f'Kill {pid} on {hostname}')
                        subprocess.run(kill_cmd, shell=True)

                    ps_pids = self.get_remote_pids(hostname=hostname, cmd=self.get_resolved_value())
            except subprocess.CalledProcessError:
                self.env.LOG.warning('Could not check for remaining remote processes.')

            self.env.LOG.info('Keystroke interrupt handled.\n')

        if self.env.imports:
            cleanup_cmd = f'{ssh_cmd} rm -r {self.env.config["remote_tmp_dir"]}'
            self.env.LOG.debug(f'Executing: {cleanup_cmd}')
            proc = subprocess.run(cleanup_cmd, shell=True, executable='/bin/bash')
            if proc.returncode != 0:
                self.env.LOG.warning(
                    'Failed to remove {tmp_dir}, exitcode: {return_code}'.format(
                        tmp_dir=self.env.config["remote_tmp_dir"],
                        return_code=proc.returncode,
                    )
                )

        return exitcode

    def _build_command(self, path: str) -> str:
        if not self.env.imports:
            return self.get_resolved_value()
        return f'. {path}/' + f'; . {path}/'.join(self.env.imports) + '; ' + self.get_resolved_value()

    def _run_local_command(self, cmd: str) -> int:
        self.env.LOG.debug(f'Executing: {cmd}')
        if self.assignment_var:
            proc = subprocess.run(cmd, shell=True, executable='/bin/bash', stdout=subprocess.PIPE)
            output = proc.stdout.decode(self.env.config["encoding"])
            self.env.vars[self.assignment_var] = output
            self.env.LOG.info(f'Variable {self.assignment_var} = {output}')
        else:
            proc = subprocess.run(cmd, shell=True, executable='/bin/bash')
        return proc.returncode

    def get_remote_pids(self, hostname, cmd) -> []:
        ps_cmd = f"ps axu | grep {quote(cmd)} | grep -v 'grep' | awk '{{print $2}}'"
        cmd = f'ssh {hostname} {quote(ps_cmd)} 2>&1'
        pids = subprocess.check_output(
            cmd,
            shell=True,
            executable='/bin/bash'
        ).decode(self.env.config["encoding"]).split()

        return pids


def parse_key(key) -> Tuple[str, ...]:
    """
    parses the key

    returns a list containing:
    - condition_var: if this evolves to False, the command is not executed
    - assignment_var: name of a variable, to which the command output is assigned
    - command_type
    """

    return re.search(r'((.*)\?)?((.*)=)?(.*)', key).group(2, 4, 5)


class AbortException(Exception):
    def __init__(self, return_code: int):
        self.return_code = return_code

    def __int__(self):
        return self.return_code


class UnknownCommandException(Exception):
    pass



from .command import Command


class BWCommand(Command):
    def _generate_python_vars(self):
        locale_vars = {}
        for key, value in self.env.systems.items():
            locale_vars[f'{key}_node'] = self.env.config["bw_repo"].get_node(value)
        locale_vars['vars'] = self.env.vars
        return locale_vars

    def _get_remote_hostname(self):
        node = self.env.config['bw_repo'].get_node(self.get_system())
        return node.hostname



from subprocess import CalledProcessError

import pytest

from automatix.command import Command, parse_key
from tests.test_environment import environment, run_command_and_check, ssh_up

pytest_plugins = ["docker_compose"]


def test__execute_remote_cmd(ssh_up):
    cmd = Command(pipeline_cmd={'remote@testsystem': 'touch /test_remote_cmd'}, index=2, env=environment)
    cmd.execute()
    try:
        run_command_and_check('ssh docker-test ls /test_remote_cmd >/dev/null')
    except CalledProcessError:
        pytest.fail('Check for remote file not successful')


def test__execute_local_cmd(capfd):
    test_string = "Local Test String"

    # empty captured stdin and stderr
    _ = capfd.readouterr()

    cmd = Command(pipeline_cmd={'local': f'echo {test_string}'}, index=2, env=environment)
    cmd.execute()

    out, err = capfd.readouterr()
    assert test_string in out
    assert err == ''


def test__execute_local_with_condition(capfd):
    condition_tests = {
        'false_var': False,
        'true_var': True,
        'none_var': False,
        'empty_var': False,
        'example_string': True,
    }

    for condition_var, condition_out in condition_tests.items():
        # empty captured stdin and stderr
        _ = capfd.readouterr()

        cmd = Command(pipeline_cmd={f'{condition_var}?local': 'pwd'}, index=2, env=environment)
        cmd.execute()

        out, err = capfd.readouterr()
        assert condition_out == ('automatix' in out)
        assert err == ''


def test__execute_python_cmd():
    test_cmd = """
from uuid import uuid4
from pprint import pprint
PERSISTENT_VARS.update(locals())
"""

    cmd = Command(pipeline_cmd={'python': test_cmd}, index=2, env=environment)
    cmd.execute()

    cmd = Command(pipeline_cmd={'python': 'print(uuid4())'}, index=2, env=environment)
    cmd.execute()


def test__parse_key():
    assert parse_key('python') == (None, None, 'python')
    assert parse_key('remote@v1') == (None, None, 'remote@v1')
    assert parse_key('host=remote@v1') == (None, 'host', 'remote@v1')
    assert parse_key('is_jira?host=remote@v1') == ('is_jira', 'host', 'remote@v1')
    assert parse_key('is_jira?python') == ('is_jira', None, 'python')



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='automatix',
    version='1.3.0',
    description='Automation wrapper for bash and python commands',
    keywords=['bash', 'shell', 'command', 'automation', 'process', 'wrapper', 'devops', 'system administration'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/seibert-media/automatix',
    author='Johannes Paul, //SEIBERT/MEDIA GmbH',
    author_email='jpaul@seibert-media.net',
    license='MIT',
    python_requires='>=3.6',
    install_requires=[
        'pyyaml>=5.1',
    ],
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'automatix=automatix:main',
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
)



import logging

from argparse import Namespace
from collections import OrderedDict
from typing import List

from .command import Command, AbortException
from .environment import PipelineEnvironment


class Automatix:
    def __init__(self, script: dict, variables: dict, config: dict, cmd_class: type, script_fields: OrderedDict):
        self.script = script
        self.script_fields = script_fields
        self.cmdClass = cmd_class
        self.env = PipelineEnvironment(
            name=script['name'],
            config=config,
            systems=script.get('systems', {}),
            variables=variables,
            imports=script.get('imports', []),
            LOG=logging.getLogger(config['logger']),
        )

    def build_command_list(self, pipeline: str) -> List[Command]:
        command_list = []
        for index, cmd in enumerate(self.script[pipeline]):
            new_cmd = self.cmdClass(
                pipeline_cmd=cmd,
                index=index,
                env=self.env,
            )
            command_list.append(new_cmd)
            if new_cmd.assignment_var and new_cmd.assignment_var not in self.env.vars:
                self.env.vars[new_cmd.assignment_var] = f'{{{new_cmd.assignment_var}}}'
        return command_list

    def print_main_data(self):
        self.env.LOG.info('\n\n')
        self.env.LOG.info('//////////////////////////////////////////////////////////////////////')
        self.env.LOG.info(f"---- {self.script['name']} ----")
        self.env.LOG.info('//////////////////////////////////////////////////////////////////////')
        for field_key, field_value in self.script_fields.items():
            self.env.LOG.info(f'\n{field_value}:')
            for key, value in self.script.get(field_key, {}).items():
                self.env.LOG.info(f" {key}: {value}")

    def print_command_line_steps(self, command_list: List[Command]):
        self.env.LOG.info('\nCommandline Steps:')
        for cmd in command_list:
            self.env.LOG.info(f"({cmd.index}) [{cmd.orig_key}]: {cmd.get_resolved_value()}")

    def execute_pipeline(self, command_list: List[Command], args: Namespace, start_index: int = 0):
        for cmd in command_list[start_index:]:
            cmd.execute(interactive=args.interactive, force=args.force)

    def execute_extra_pipeline(self, pipeline: str):
        try:
            if self.script.get(pipeline):
                pipeline_list = self.build_command_list(pipeline=pipeline)
                self.env.LOG.info('\n------------------------------')
                self.env.LOG.info(f' --- Start {pipeline.upper()} pipeline ---')
                self.execute_pipeline(command_list=pipeline_list, args=Namespace(interactive=False, force=False))
                self.env.LOG.info(f'\n --- End {pipeline.upper()} pipeline ---')
                self.env.LOG.info('------------------------------\n')
        except AbortException as exc:
            exit(int(exc))
        except KeyboardInterrupt:
            self.env.LOG.warning('\nAborted by user. Exiting.')
            exit(130)

    def run(self, args: Namespace):
        command_list = self.build_command_list(pipeline='pipeline')

        self.execute_extra_pipeline(pipeline='always')

        self.print_main_data()
        self.print_command_line_steps(command_list)
        if args.print_overview:
            exit()

        try:
            self.execute_pipeline(command_list=command_list, args=args, start_index=int(args.jump_to))
        except AbortException as exc:
            self.env.LOG.debug('Abort requested. Cleaning up.')
            self.execute_extra_pipeline(pipeline='cleanup')
            self.env.LOG.debug('Clean up done. Exiting.')
            exit(int(exc))
        except KeyboardInterrupt:
            self.env.LOG.warning('\nAborted by user. Exiting.')
            exit(130)

        self.execute_extra_pipeline(pipeline='cleanup')

        self.env.LOG.info('---------------------------------------------------------------')
        self.env.LOG.info('Automatix finished: Congratulations and have a N.I.C.E. day :-)')
        self.env.LOG.info('---------------------------------------------------------------')
